@@ -20,7 +20,7 @@ from tf_bind_transformer.cache_utils import cache_fn
 from tf_bind_transformer.protein_utils import get_protein_embedder
 from tf_bind_transformer.context_utils import get_text_repr, get_contextual_dim
 
-from tf_bind_transformer.attention import FeedForward, JointCrossAttention
+from tf_bind_transformer.attention import FeedForward, JointCrossAttentionBlock, CrossAttention
 
 # helper functions
 
@@ -160,8 +160,98 @@ class ReadValueMLP(nn.Module):
 
         return F.mse_loss(pred, read_value)
 
+class HypergridLinear(nn.Module):
+    def __init__(
+        self,
+        dim,
+        dim_out,
+        *,
+        context_dim
+    ):
+        super().__init__()
+        self.weights = nn.Parameter(torch.randn(dim, dim_out))
+        self.contextual_projection = nn.Linear(context_dim, dim * dim_out)
+
+    def forward(self, x, context):
+        # derive contextual gating, from hypergrids paper
+
+        gating = self.contextual_projection(context).sigmoid()
+        gating = rearrange(gating, 'b (i o) -> b i o', i = int(math.sqrt(gating.shape[-1])))
+
+        # gate interactions projection with context
+
+        to_logits_w = rearrange(self.weights, 'i o -> 1 i o') * gating
+        return einsum('b n d, b d e -> b n e', x, to_logits_w)
+
 # FILIP adapter model
 
+class FILIP(nn.Module):
+    def __init__(
+        self,
+        dim,
+        context_dim,
+        heads,
+        dim_head = 64,
+        dropout = 0.
+    ):
+        super().__init__()
+        self.heads = heads
+        inner_latent_dim = heads * dim_head
+
+        self.to_latent_w = nn.Parameter(torch.randn(dim, inner_latent_dim))
+        self.to_latent_b = nn.Parameter(torch.randn(inner_latent_dim))
+
+        self.context_to_latent_w = nn.Parameter(torch.randn(context_dim, inner_latent_dim))
+        self.context_to_latent_b = nn.Parameter(torch.randn(inner_latent_dim))
+
+        self.dropout = nn.Dropout(dropout)
+
+    def forward(
+        self,
+        x,
+        context,
+        context_mask = None
+    ):
+        heads = self.heads
+
+        x = einsum('b n d, d e -> b n e', x, self.to_latent_w)
+        x = x + self.to_latent_b
+
+        x = rearrange(x, 'b n (h d) -> b h n d', h = heads)
+
+        context = einsum('b n d, d e -> b n e', context, self.context_to_latent_w)
+        context = context + self.context_to_latent_b
+
+        context = rearrange(context, 'b n (h d) -> b h n d', h = heads)
+
+        context, x = map(l2norm, (context, x))
+
+        # fine grained interaction between dna and protein sequences
+        # FILIP https://arxiv.org/abs/2111.07783
+
+        if x.shape[0] == 1:
+            # in the case one passes in 1 genomic sequence track
+            # but multiple factors + contexts, as in enformer training
+            x = rearrange(x, '1 ... -> ...')
+            einsum_eq = 'h i d, b h j d -> b h i j'
+        else:
+            einsum_eq = 'b h i d, b h j d -> b h i j'
+
+        interactions = einsum(einsum_eq, x, context)
+
+        # dropout
+
+        interactions = self.dropout(interactions)
+
+        # reduction
+
+        if exists(context_mask):
+            context_mask = rearrange(context_mask, 'b j -> b 1 1 j')
+
+        interactions = logavgexp(interactions, mask = context_mask, dim = -1)
+        interactions = rearrange(interactions, 'b h i -> b i h')
+        return interactions
+
 class AdapterModel(nn.Module):
     def __init__(
         self,
@@ -182,6 +272,9 @@ class AdapterModel(nn.Module):
         aux_read_value_loss = False,
         read_value_aux_loss_weight = 0.1,
         fourier_dims = 256,
+        condition_squeeze_excite = False,
+        condition_film = False,
+        condition_hypergrid = False,
         **kwargs
     ):
         super().__init__()
@@ -209,21 +302,41 @@ class AdapterModel(nn.Module):
         else:
             assert exists(aa_embed_dim), 'AA embedding dimensions must be set if not using ESM'
 
+        # conditioning
+
+        self.cond_genetic = None
+        self.cond_protein = None
+
+        if condition_squeeze_excite or condition_film:
+            condition_klass = SqueezeExcitation if condition_squeeze_excite else FiLM
+
+            self.cond_genetic  = condition_klass(contextual_embed_dim, enformer_dim)
+            self.cond_protein  = condition_klass(contextual_embed_dim, aa_embed_dim)
+
+        # joint attn
+
+        self.joint_cross_attn = JointCrossAttentionBlock(
+            dim = enformer_dim,
+            context_dim = aa_embed_dim,
+            dropout = dropout
+        )
+
         # latents
 
-        self.latent_heads = latent_heads
-        inner_latent_dim = latent_heads * latent_dim
+        self.filip = FILIP(
+            dim = enformer_dim,
+            context_dim = aa_embed_dim,
+            dim_head = latent_dim,
+            heads = latent_heads,
+            dropout = dropout
+        )
 
-        self.seq_embed_to_latent_w = nn.Parameter(torch.randn(enformer_dim, inner_latent_dim))
-        self.seq_embed_to_latent_b = nn.Parameter(torch.randn(inner_latent_dim))
+        # hypergrid conditioning
 
-        self.aa_seq_embed_to_latent_w = nn.Parameter(torch.randn(aa_embed_dim, inner_latent_dim))
-        self.aa_seq_embed_to_latent_b = nn.Parameter(torch.randn(inner_latent_dim))
+        self.linear_with_hypergrid = None
 
-        self.to_logits_w = nn.Parameter(torch.randn(latent_heads, latent_heads))
-        self.contextual_projection = nn.Linear(contextual_embed_dim, latent_heads * latent_heads)
-
-        self.dropout = nn.Dropout(dropout)
+        if exists(condition_hypergrid):
+            self.linear_with_hypergrid = HypergridLinear(latent_heads, latent_heads, context_dim = contextual_embed_dim)
 
         # to prediction
 
@@ -276,7 +389,6 @@ class AdapterModel(nn.Module):
         finetune_enformer_ln_only = False
     ):
         device = seq.device
-        latent_heads = self.latent_heads
 
         # prepare enformer for training
         # - set to eval and no_grad if not fine-tuning
@@ -323,304 +435,35 @@ class AdapterModel(nn.Module):
                 device = seq.device
             )
 
+        # contextual conditioning
+        # film or squeeze-excite for both genetic and protein sequences
+
+        if exists(self.cond_genetic):
+            seq_embed = self.cond_genetic(seq_embed, contextual_embed)
+
+        if exists(self.cond_protein):
+            aa_embed = self.cond_protein(aa_embed, contextual_embed, mask = aa_mask)
+
+        # joint cross attention
+
+        seq_embed, aa_embed = self.joint_cross_attn(
+            seq_embed,
+            context = aa_embed,
+            context_mask = aa_mask
+        )
+
         # project both embeddings into shared latent space
 
-        seq_latent = einsum('b n d, d e -> b n e', seq_embed, self.seq_embed_to_latent_w)
-        seq_latent = seq_latent + self.seq_embed_to_latent_b
+        interactions = self.filip(
+            seq_embed,
+            aa_embed,
+            context_mask = aa_mask
+        )
 
-        seq_latent = rearrange(seq_latent, 'b n (h d) -> b h n d', h = latent_heads)
 
-        aa_latent = einsum('b n d, d e -> b n e', aa_embed, self.aa_seq_embed_to_latent_w)
-        aa_latent = aa_latent + self.aa_seq_embed_to_latent_b
+        # linear with hypergrid conditioning
 
-        aa_latent = rearrange(aa_latent, 'b n (h d) -> b h n d', h = latent_heads)
-
-        aa_latent, seq_latent = map(l2norm, (aa_latent, seq_latent))
-
-        # fine grained interaction between dna and protein sequences
-        # FILIP https://arxiv.org/abs/2111.07783
-
-        if seq_latent.shape[0] == 1:
-            # in the case one passes in 1 genomic sequence track
-            # but multiple factors + contexts, as in enformer training
-            seq_latent = rearrange(seq_latent, '1 ... -> ...')
-            einsum_eq = 'h i d, b h j d -> b h i j'
-        else:
-            einsum_eq = 'b h i d, b h j d -> b h i j'
-
-        interactions = einsum(einsum_eq, seq_latent, aa_latent)
-
-        # dropout
-
-        interactions = self.dropout(interactions)
-
-        # reduction
-
-        if exists(aa_mask):
-            aa_mask = rearrange(aa_mask, 'b j -> b 1 1 j')
-
-        interactions = logavgexp(interactions, mask = aa_mask, dim = -1)
-        interactions = rearrange(interactions, 'b h i -> b i h')
-
-        # derive contextual gating, from hypergrids paper
-
-        gating = self.contextual_projection(contextual_embed).sigmoid()
-        gating = rearrange(gating, 'b (i o) -> b i o', i = int(math.sqrt(gating.shape[-1])))
-
-        # gate interactions projection with context
-
-        to_logits_w = rearrange(self.to_logits_w, 'i o -> 1 i o') * gating
-        logits = einsum('b n d, b d e -> b n e', interactions, to_logits_w)
-
-        # to *-seq prediction
-
-        pred = self.to_pred(logits)
-
-        if not exists(target):
-            return pred
-
-        if exists(target) and return_corr_coef:
-            return pearson_corr_coef(pred, target)
-
-        if exists(target) and not self.binary_target:
-            return poisson_loss(pred, target)
-
-        if not exists(target):
-            return pred
-
-        # binary loss w/ optional auxiliary loss
-
-        loss = self.loss_fn(pred, target.float())
-
-        if not self.aux_read_value_loss:
-            return loss, torch.Tensor([0.]).to(device)
-
-        # return prediction if not auto-calculating loss
-
-        assert exists(read_value) and exists(peaks_nr), 'peaks NR must be supplied if doing auxiliary read value loss'
-
-        aux_loss = self.to_read_value_aux_loss(
-            logits,
-            peaks_nr,
-            read_value = read_value
-        )
-
-        return loss, aux_loss
-
-
-# cross attention based tf-bind-transformer
-
-class AttentionAdapterModel(nn.Module):
-    def __init__(
-        self,
-        *,
-        enformer,
-        aa_embed_dim = None,
-        aa_embed_encoder = 'esm',
-        contextual_embed_dim = None,
-        cross_attn_dim_head = 64,
-        cross_attn_heads = 8,
-        use_aa_embeds = False,
-
-        use_free_text_context = False,
-        free_text_context_encoder = 'pubmed',
-        free_text_embed_method = 'cls',
-        dropout = 0.,
-        use_squeeze_excite = False,
-        binary_target = False,
-        target_mse_loss = False,
-        aux_read_value_loss = False,
-        read_value_aux_loss_weight = 0.1,
-        fourier_dims = 256
-    ):
-        super().__init__()
-        assert isinstance(enformer, Enformer), 'enformer must be an instance of Enformer'
-        self.enformer = enformer
-        enformer_dim = enformer.dim * 2
-
-        self.norm_seq_embed = nn.LayerNorm(enformer_dim)
-
-        # contextual embedding related variables
-
-        assert free_text_embed_method in {'cls', 'mean_pool'}, 'must be either cls or mean_pool'
-        self.free_text_embed_method = free_text_embed_method
-        self.use_free_text_context = use_free_text_context
-        contextual_embed_dim = default(contextual_embed_dim, get_contextual_dim(free_text_context_encoder))
-
-        # protein embedding related variables
-
-        self.use_aa_embeds = use_aa_embeds
-        self.aa_embed_config = get_protein_embedder(aa_embed_encoder)
-        self.get_aa_embed = self.aa_embed_config['fn']
-
-        if use_aa_embeds:
-            aa_embed_dim = self.aa_embed_config['dim']
-        else:
-            assert exists(aa_embed_dim), 'AA embedding dimensions must be set if not using ESM'
-
-        # film
-
-        condition_klass = SqueezeExcitation if use_squeeze_excite else FiLM
-
-        self.cond_genetic  = condition_klass(contextual_embed_dim, enformer_dim)
-        self.cond_genetic2 = condition_klass(contextual_embed_dim, enformer_dim)
-        self.cond_protein  = condition_klass(contextual_embed_dim, aa_embed_dim)
-
-        # cross attention
-
-        self.attn_prenorm = nn.LayerNorm(enformer_dim)
-
-        self.scale = cross_attn_dim_head ** -0.5
-        self.heads = cross_attn_heads
-        inner_dim = cross_attn_dim_head * cross_attn_heads
-
-        self.to_queries = nn.Linear(enformer_dim, inner_dim, bias = False)
-        self.to_keys = nn.Linear(aa_embed_dim, inner_dim, bias = False)
-        self.to_values = nn.Linear(aa_embed_dim, inner_dim, bias = False)
-        self.to_out = nn.Linear(inner_dim, enformer_dim)
-
-        self.attn_dropout = nn.Dropout(dropout)
-
-        self.feedforward = FeedForward(enformer_dim, dropout = dropout)
-
-        # to predictions
-
-        self.binary_target = binary_target
-        self.aux_read_value_loss = aux_read_value_loss
-        self.read_value_aux_loss_weight = read_value_aux_loss_weight
-
-        if binary_target:
-            self.loss_fn = F.binary_cross_entropy_with_logits if not target_mse_loss else F.mse_loss
-
-            self.to_pred = nn.Sequential(
-                Reduce('... n d -> ... d', 'mean'),
-                nn.LayerNorm(enformer_dim),
-                nn.Linear(enformer_dim, 1),
-                Rearrange('... 1 -> ...')
-            )
-
-            self.to_read_value_aux_loss = ReadValueMLP(
-                dim = enformer_dim,
-                fourier_dims = fourier_dims
-            )
-        else:
-            self.to_pred = nn.Sequential(
-                nn.Linear(enformer_dim, 1),
-                Rearrange('... 1 -> ...'),
-                nn.Softplus()
-            )
-
-    def combine_losses(self, loss, aux_loss):
-        if not self.aux_read_value_loss:
-            return loss
-
-        return loss + self.read_value_aux_loss_weight * aux_loss
-
-    def forward(
-        self,
-        seq,
-        *,
-        aa = None,
-        aa_embed = None,
-        contextual_embed = None,
-        contextual_free_text = None,
-        aa_mask = None,
-        target = None,
-        read_value = None,
-        peaks_nr = None,
-        return_corr_coef = False,
-        finetune_enformer = False,
-        finetune_enformer_ln_only = False
-    ):
-        device = seq.device
-        h = self.heads
-
-        # prepare enformer for training
-        # - set to eval and no_grad if not fine-tuning
-        # - always freeze the batchnorms
-
-        freeze_batchnorms_(self.enformer)
-        enformer_forward = self.enformer.forward
-
-        if finetune_enformer:
-            enformer_context = null_context()
-        elif finetune_enformer_ln_only:
-            enformer_context = null_context()
-            freeze_all_but_layernorms_(self.enformer)
-        else:
-            self.enformer.eval()
-            enformer_context = torch.no_grad()
-            enformer_forward_wrapper = cache_enformer_forward if self.training else identity
-            enformer_forward = enformer_forward_wrapper(enformer_forward)
-
-        # genetic sequence embedding
-
-        with enformer_context:
-            seq_embed = enformer_forward(seq, return_only_embeddings = True)
-
-        seq_embed = self.norm_seq_embed(seq_embed)
-
-        # protein related embeddings
-
-        if self.use_aa_embeds:
-            assert exists(aa), 'aa must be passed in as tensor of integers from 0 - 20 (20 being padding)'
-            aa_embed, aa_mask = self.get_aa_embed(aa, device = seq.device)
-        else:
-            assert exists(aa_embed), 'protein embeddings must be given as aa_embed'
-
-        # free text embeddings, for cell types and experimental params
-
-        if not exists(contextual_embed):
-            assert self.use_free_text_context, 'use_free_text_context must be set to True if one is not passing in contextual_embed tensor'
-            assert exists(contextual_free_text), 'context must be supplied as array of strings as contextual_free_text if contextual_embed is not supplied'
-
-            contextual_embed = get_text_repr(
-                contextual_free_text,
-                return_cls_token = (self.free_text_embed_method == 'cls'),
-                device = seq.device
-            )
-
-        # in the case that a single genetic sequence was given, but needs to interact with multiple contexts
-
-        num_seq = seq_embed.shape[0]
-        num_contexts = aa_embed.shape[0]
-
-        if num_seq == 1:
-            seq_embed = repeat(seq_embed, '1 n d -> b n d', b = num_contexts)
-
-        # film condition both genetic and protein sequences
-
-        seq_embed = self.cond_genetic(seq_embed, contextual_embed)
-        aa_embed = self.cond_protein(aa_embed, contextual_embed, mask = aa_mask)
-
-        # cross attention
-
-        normed_seq_embed = self.attn_prenorm(seq_embed)
-        queries, keys, values = self.to_queries(normed_seq_embed), self.to_keys(aa_embed), self.to_values(aa_embed)
-        queries, keys, values = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), (queries, keys, values))
-
-        sim = einsum('b h i d, b h j d -> b h i j', queries, keys) * self.scale
-
-        aa_mask = rearrange(aa_mask, 'b j -> b 1 1 j')
-
-        sim = sim.masked_fill(~aa_mask, -torch.finfo(sim.dtype).max)
-
-        attn = sim.softmax(dim = -1)
-        attn = self.attn_dropout(attn)
-
-        out = einsum('b h i j, b h j d -> b h i d', attn, values)
-        out = rearrange(out, 'b h n d -> b n (h d)')
-        out = self.to_out(out)
-
-        seq_embed = seq_embed + out
-
-        # condition one more time
-
-        seq_embed = self.cond_genetic2(seq_embed, contextual_embed)
-
-        # feedforward
-
-        logits = self.feedforward(seq_embed) + seq_embed
+        logits = self.linear_with_hypergrid(interactions, context = contextual_embed)
 
         # to *-seq prediction
 
